import asyncio
from collections.abc import Callable
from pprint import pprint

import pop.hub


def main():
    # Create the hub, loading all dynes and starting the loop
    hub = pop.hub.Hub(cli="pop_cli")
    finder = hub

    ref = hub.OPT.pop_cli.ref
    parts = ref.split(".")
    for p in parts:
        if not p:
            continue
        finder = getattr(finder, p)

    args = []
    kwargs = {}

    # Override hub.OPT with the the new cli
    cli = hub.OPT.pop_cli.cli
    if cli:
        # Pass all remaining args onto the new parser
        hub.OPT = hub._sync.pop.config.load(cli=cli, parser_args=hub.OPT.pop_cli.args)
    else:
        # We are using the pop cli, treat all the extra args as parameters for the called function
        PLACEHOLDER = object()
        hold = PLACEHOLDER
        for arg in hub.OPT.pop_cli.args:
            if hold is not PLACEHOLDER:
                key = hold.replace("-", "_").lstrip("_")
                if "," in arg:
                    arg = arg.split(",")
                kwargs[key] = arg
                hold = PLACEHOLDER
            elif "=" in arg:
                key, value = arg.split("=", maxsplit=1)
                key = key.replace("-", "_").lstrip("_")
                if "," in value:
                    value = value.split(",")
                kwargs[key] = value
            elif arg.startswith("--"):
                hold = arg
            else:
                args.append(arg)

    # Call the named reference on the hub
    # This allows you to do
    # $ pop idem.init.cli
    # This way you can have multiple entrypoints, or even alias the above command to "idem"
    if asyncio.iscoroutinefunction(finder):
        pprint(hub._synchronize(finder))
    elif isinstance(finder, Callable):
        ret = finder(*args, **kwargs)
        while asyncio.iscoroutine(ret):
            ret = hub._synchronize(ret)

        pprint(ret)
    else:
        pprint(finder)


if __name__ == "__main__":
    main()



async def func(hub, *args, **kwargs):
    return args, kwargs


async def benchmark(hub, depth=5, width=10000):
    """
    Create a bloated hub

    Run a pop with all the strict contract checking

    .. code-block:: bash

        python -m hub pop.test.benchmark

    Run an optimized pop

    .. code-block:: bash

        python -O -m hub pop.test.benchmark
    """
    await hub.log.debug("Started Benchmark")
    start_time = hub.lib.time.time()
    for w in range(int(width)):
        name = f"sub_{w}"
        await hub.pop.sub.add(dyne_name=name)
        dyne = hub[name]
        for d in range(int(depth)):
            nest_name = f"sub_{d}_{d}"
            await hub.pop.sub.add(dyne_name=nest_name, sub=dyne)
            dyne = dyne[nest_name]
    end_time = hub.lib.time.time()
    return f"Total time taken: {end_time - start_time:.2f} seconds."



import inspect
import sys

import pop.contract


def _stack_frames(relative_start: int) -> inspect.FrameInfo:
    """
    Efficiently access stack frames.
    :param relative_start: Starting stack depth; The default, 2 is the parent of the
            caller of stack_frames - the first function that may be unknown.
    :return: a stack frame
    """
    if hasattr(sys, "_getframe"):
        # implementation detail of CPython, speeds things up by 100x.
        frame = sys._getframe(relative_start)
        while frame:
            yield frame
            frame = frame.f_back
    else:
        for frame_info in inspect.stack(context=0)[relative_start:]:
            yield frame_info.frame


def _get_hub_ref():
    # Minimize lookup time by starting at frame 5, it will be at least that far back
    for frame in _stack_frames(5):
        if isinstance(frame.f_locals.get("self"), pop.contract.Contracted):
            contracted = frame.f_locals["self"]
            return contracted, frame.f_lineno

    # Default to the root reference
    return None, 0


def _get_logger(hub, name: str = ""):
    if name not in hub.log.LOGGER:
        hub.log.LOGGER[name] = hub.lib.aiologger.Logger(name=name)
        hub.log.LOGGER[name].level = hub.log.INT_LEVEL
        if hub.log.FILE_HANDLER:
            hub.log.LOGGER[name].handlers.append(hub.log.FILE_HANDLER)
        if hub.log.STREAM_HANDLER:
            hub.log.LOGGER[name].handlers.append(hub.log.STREAM_HANDLER)
    return hub.log.LOGGER[name]


async def log(hub, level: int, msg: str, *args, **kwargs):
    if hub.log.INT_LEVEL > level:
        return
    contract, lineno = _get_hub_ref()

    if contract:
        caller = f"{contract.ref}.{contract.func.__name__}"
        mod = contract.func.__module__
        func = contract.func.__name__
    else:
        caller = "hub"
        mod = "hub"
        func = "hub"

    logger = _get_logger(hub, caller)
    record = hub.lib.aiologger.records.LogRecord(
        name=caller,
        pathname=mod,
        lineno=lineno,
        level=0,  # We have to overwrite this in a secure way
        msg=msg,
        args=args,
        func=func,
        **kwargs,
    )
    record.levelno = level
    if level == 5:
        record.levelname = "TRACE"
    else:
        try:
            record.levelname = hub.lib.aiologger.records.get_level_name(level)
        except ValueError:
            record.levelname = f"LEVEL {level}"

    await logger.handle(record)


async def setup(hub, conf: dict[str, any]):
    """
    Given the configuration data set up the logger.
    """
    # Set the logging level
    raw_level = conf["log_level"].strip().lower()
    hub.log.INT_LEVEL = (
        int(raw_level)
        if raw_level.isdigit()
        else hub.log.LEVEL.get(raw_level, hub.log.LEVEL["info"])
    )

    # Get the root logger
    root = _get_logger(hub, name="")
    root.level = hub.log.INT_LEVEL

    # Set up the console handler
    console_formatter = hub.lib.aiologger.formatters.base.Formatter(
        fmt=conf["log_fmt_console"], datefmt=conf["log_datefmt"]
    )
    console_handler = hub.lib.aiologger.handlers.streams.AsyncStreamHandler(
        formatter=console_formatter, stream=hub.lib.sys.stderr
    )
    root.add_handler(console_handler)
    hub.log.STREAM_HANDLER = console_handler

    # Set up the file handler
    log_file = hub.lib.pathlib.Path(conf["log_file"]).expanduser()
    log_file.parent.mkdir(parents=True, exist_ok=True)

    file_formatter = hub.lib.aiologger.formatters.base.Formatter(
        fmt=conf["log_fmt_logfile"], datefmt=conf["log_datefmt"]
    )
    file_handler = hub.lib.aiologger.handlers.files.AsyncFileHandler(log_file)
    file_handler.formatter = file_formatter
    root.add_handler(file_handler)
    hub.log.FILE_HANDLER = file_handler

    # Put all these functions higher up on the hub
    hub.log.log = hub.log["async"].log
    hub.log.trace = lambda msg, *args, **kwargs: hub.log.log(
        level=5, msg=msg, *args, **kwargs
    )
    hub.log.debug = lambda msg, *args, **kwargs: hub.log.log(
        level=hub.lib.aiologger.levels.LogLevel.DEBUG, msg=msg, *args, **kwargs
    )
    hub.log.info = lambda msg, *args, **kwargs: hub.log.log(
        level=hub.lib.aiologger.levels.LogLevel.INFO, msg=msg, *args, **kwargs
    )
    hub.log.warning = lambda msg, *args, **kwargs: hub.log.log(
        level=hub.lib.aiologger.levels.LogLevel.WARNING, msg=msg, *args, **kwargs
    )
    hub.log.error = lambda msg, *args, **kwargs: hub.log.log(
        level=hub.lib.aiologger.levels.LogLevel.ERROR, msg=msg, *args, **kwargs
    )
    hub.log.critical = lambda msg, *args, **kwargs: hub.log.log(
        level=hub.lib.aiologger.levels.LogLevel.CRITICAL, msg=msg, *args, **kwargs
    )



"""
Pop related exceptions
"""


class PopBaseException(Exception):
    """
    Base exception where all of Pop's exceptions derive
    """


class PopError(PopBaseException):
    """
    General purpose pop exception to signal an error
    """


class PopLoadError(PopBaseException):
    """
    Exception raised when a pop module fails to load
    """


class PopLookupError(PopBaseException):
    """
    Exception raised when a pop module lookup fails
    """


class ContractModuleException(PopBaseException):
    """
    Exception raised when a function specified in a contract as required
    to exist is not found in the loaded module
    """


class ContractFuncException(PopBaseException):
    """
    Exception raised when a function specified in a contract as required
    to exist is found on the module but it's not function
    """


class ContractSigException(PopBaseException):
    """
    Exception raised when a function signature is not compatible with the
    coresponding function signature found in the contract.
    """


class ProcessNotStarted(PopBaseException):
    """
    Exception raised when failing to start a process on the process manager
    """


class BindError(PopBaseException):
    """
    Exception raised when arguments for a function in a ContractedContext cannot be bound
    Indicates invalid function arguments.
    """


class PreContractFailed(PopBaseException):
    """
    Exception raised when a pre contract returns False or False and a message.
    Indicates that the corresponding function could not be called.
    """



"""
Used to resolve resolutions to paths on the hub
"""


def last(hub: "pop.hub.Hub", ref: str) -> "pop.hub.Sub":
    """
    Takes a string that references the desired ref and returns the last object
    called out in that ref
    """
    return path(hub, ref)[-1]


def path(hub: "pop.hub.Hub", ref: str) -> list["pop.hub.Sub"]:
    """
    Retuns a list of references up to the named ref
    """
    ret = [hub]
    if isinstance(ref, str):
        ref = ref.split(".")
    for chunk in ref:
        ret.append(getattr(ret[-1], chunk))
    return ret



from collections.abc import Callable

import pop.exc


async def process_pre_result(
    hub, pre_ret, pre: Callable, cn: "pop.contract.Contracted"
):
    if pre_ret is None:
        return

    if isinstance(pre_ret, bool):
        status = pre_ret
        msg = None

    if isinstance(pre_ret, tuple):
        status, msg = pre_ret

    if not msg:
        msg = f"Pre contract '{pre.__name__}' failed for function '{cn.func.__name__}'"

    # If the return of "pre" is "False" then raise an error
    if status is False:
        raise pop.exc.PreContractFailed(msg)



"""
Load the files detected from the scanner
"""
import asyncio
import importlib.machinery
import importlib.util
import inspect
import os
import re
import sys
import types
from collections.abc import Callable
from typing import Any
from typing import List

import pop.contract
import pop.data
import pop.exc


class LoadError(Exception):
    """
    Errors from the loader are contained herein
    """

    __slots__ = ("edict", "traceback")

    def __init__(self, msg, exception=None, traceback=None, verror=None):
        self.edict = {
            "msg": msg,
            "exception": exception,
            "verror": verror,
        }
        self.traceback = traceback

    def __call__(self):
        """
        Return the error cases
        """
        return self.edict

    def __getattr__(self, attr):
        """
        Return a lambda that returns the edict
        """
        return self.__calling_load_error__


def load_mod(modname: str, form: str, path: str) -> "LoadedMod":
    """
    Load a single module
    :param modname: The name of the module to get from the loader
    :param form: The name of the loader module
    :param path: The package to use as the anchor point from which to resolve the
        relative import to an absolute import.
    """
    this = sys.modules[__name__]
    return getattr(this, form)(modname, path)


def python(modname: str, path: str) -> "LoadedMod" or LoadError:
    """
    Attempt to load the named python modules
    :param modname: The name of the module to get from the loader
    :param path: The package to use as the anchor point from which to resolve the
        relative import to an absolute import.
    """
    if modname in sys.modules:
        return sys.modules[modname]
    spec = importlib.util.spec_from_file_location(modname, path)
    module = importlib.util.module_from_spec(spec)
    sys.modules[modname] = module
    module.__file__ = path
    spec.loader.exec_module(module)
    return module


def _base_name(bname: str, mod: "LoadedMod") -> tuple[str, str]:
    """
    Find the basename and alias of a loader module
    :param bname: The base name of the mod's path
    :param mod: A loader module or a LoadError if the module didn't load
    """
    base_name = os.path.basename(bname)
    return base_name, getattr(mod, "__virtualname__", base_name)


async def _load_virtual(
    hub: "pop.hub.Hub",
    mod: "LoadedMod" or LoadError,
    bname: str,
    vtype: str,
) -> dict[str, Any]:
    """
    Run the virtual function to name the module and check for all loader
    errors
    :param hub: The redistributed pop central hub
    :param virtual: Toggle whether or not to process __virtual__ functions
    :param mod: A loader module or a LoadError if the module didn't load
    :param bname: The base name of the mod's path
    :param vtype: The name of the virtual function to call on the module I.E. __virtual__ or __sub_virtual__
    """
    base_name, name = _base_name(bname, mod)

    if not hasattr(mod, vtype):
        # No __virtual__ processing is required.
        # Return the mod's name as the defined __virtualname__ if defined,
        # else, the base_name
        return {"name": name}

    try:
        vret = await getattr(mod, vtype)(hub)
    except Exception as e:
        vret = False, str(e)

    verror = vret
    if isinstance(vret, tuple):
        if len(vret) > 1:
            verror = vret[1]
        vret = vret[0]

    if vret:
        # No problems occurred, module is allowed to load
        # Return the mod's name as the defined __virtualname__ if defined,
        # else, the base_name
        return {"name": name}
    else:
        # __virtual__ explicitly disabled the loading of this module
        err = LoadError(f"Module {bname} returned virtual FALSE", verror=verror)
        # Return the load error with name as the base_name because another
        # module is still allowed to load under the same __virtualname__
        # but also return the vname information
        return {"name": base_name, "vname": name, "error": err}


def load_virtual(
    hub: "pop.hub.Hub", mod: "LoadedMod" or LoadError, bname: str
) -> dict[str, Any]:
    """
    Run the __virtual__ function to name the module and check for all loader errors
    :param hub: The redistributed pop central hub
    :param virtual: Toggle whether or not to process __virtual__ functions
    :param mod: A loader module or a LoadError if the module didn't load
    :param bname: The base name of the mod's path
    """
    return _load_virtual(hub, mod, bname, "__virtual__")


async def load_sub_virtual(
    hub: "pop.hub.Hub", mod: "LoadedMod" or LoadError, bname: str
) -> dict[str, Any]:
    """
    Run the __sub_virtual__ function to name the module and check for all loader errors
    :param hub: The redistributed pop central hub
    :param virtual: Toggle whether or not to process __virtual__ functions
    :param mod: A loader module or a LoadError if the module didn't load
    :param bname: The base name of the mod's path
    """
    _, name = _base_name(bname, mod)
    if name != "init":
        return {"name": name}
    return await _load_virtual(hub, mod, bname, "__sub_virtual__")


async def mod_init(sub: "pop.hub.Sub", mod: "LoadedMod", mod_name: str):
    """
    Process module's __init__ function if defined
    :param sub: The pop object that contains the loaded module data
    :param mod: A loader modul
    :param mod_name: The name of the module to get from the loader
    """
    if "__init__" in dir(mod):
        init = pop.contract.Contracted(
            sub._hub,
            contracts=[],
            func=mod.__init__,
            ref=f"{sub._subname}.{mod_name}",
            name="__init__",
        )
        await init()


def sub_alias(this_sub: "pop.hub.Sub", mod: "LoadedMod", mod_name: str):
    """
    Check the sub alias settings and apply the alias names locally so they can be gathered into the higher level object on the hub
    :param this_sub: The pop object that contains the loaded module data
    :param mod: A loader module
    :param mod_name: The name of the module to get from the loader
    """
    if mod_name == "init":
        alias = getattr(mod, "__sub_alias__", None)
        if alias:
            this_sub._alias = alias


async def prep_loaded_mod(
    this_sub: "pop.hub.Sub",
    mod: "LoadedMod",
    mod_name: str,
    contracts: "List[pop.contract.Wrapper]",
    recursive_contracts: "List[pop.contract.Wrapper]",
) -> "LoadedMod":
    """
    Read the attributes of a python module and create a LoadedMod, which resolves
    aliases and omits objects that should not be exposed.
    :param this_sub: The pop object that contains the loaded module data
    :param mod: A loader module
    :param mod_name: The name of the module to get from the loader
    :param contracts: Contracts functions to add to the sub
    :param recursive_contracts:
    """
    ordered_contracts = contracts + recursive_contracts

    lmod = this_sub._loaded.get(mod_name, LoadedMod(mod_name))
    ref = f"{this_sub._subname}.{mod_name}"  # getattr(hub, ref) should resolve to this module

    # Recursively get the full reference to this mod
    root = this_sub._root
    while hasattr(root, "_subname"):
        ref = f"{root._subname}.{ref}"
        root = root._root

    sub_alias(this_sub, mod, mod_name)

    __func_alias__: Callable or dict[str, str or Callable] = getattr(
        mod, "__func_alias__", {}
    )
    # If __func_alias__ is a function, run it and replace the function with it's returned dictionary
    if asyncio.iscoroutinefunction(__func_alias__):
        __func_alias__ = await __func_alias__(this_sub._hub)

    # Get all the attributes from the loaded mod
    for attr in getattr(mod, "__load__", dir(mod)):
        # For an attribute in the module, get it's function alias
        func_alias: str = __func_alias__.get(attr, attr)
        func = getattr(mod, attr)
        name = func_alias
        if not this_sub._omit_vars:
            if (
                not inspect.isfunction(func)
                and not inspect.isclass(func)
                and type(func).__name__ != "cython_function_or_method"
            ):
                lmod._vars[name] = func
                continue
        if attr.startswith(this_sub._omit_start):
            continue
        if attr.endswith(this_sub._omit_end):
            continue
        if (
            inspect.isfunction(func)
            or inspect.isbuiltin(func)
            or type(func).__name__ == "cython_function_or_method"
        ):
            obj = pop.contract.create_contracted(
                this_sub._hub,
                ordered_contracts,
                func,
                ref,
                name,
            )
            if not this_sub._omit_func:
                if this_sub._pypath and not func.__module__.startswith(mod.__name__):
                    # We're only interested in functions defined in this module, not
                    # imported functions
                    continue
                check_param_alias(obj)
                lmod._funcs[name] = obj
        else:
            klass = func
            if not this_sub._omit_class and inspect.isclass(klass):
                # We're only interested in classes defined in this module, not
                # imported classes
                if not klass.__module__.startswith(mod.__name__):
                    continue
                lmod._classes[name] = klass

    # Add external aliased functions to the loaded mod
    for attr, func_alias in __func_alias__.items():
        if isinstance(func_alias, Callable):
            obj = pop.contract.create_contracted(
                this_sub._hub,
                ordered_contracts,
                func_alias,
                ref,
                attr,
                implicit_hub=False,
            )
            lmod._funcs[attr] = obj

    # Ensure that the __verify_order__ function exists on contract modules
    verify_order: Callable = getattr(mod, "__verify_order__", lambda *_: True)
    lmod.__verify_order__ = pop.contract.create_contracted(
        this_sub._hub,
        ordered_contracts,
        verify_order,
        ref,
        attr,
        implicit_hub=True,
    )

    return lmod


def check_param_alias(func: "pop.contract.Contracted"):
    """
    Verify that parameter aliases are unique
    """
    params_and_aliases = set(func.signature.parameters.keys())
    for param, signature in func.signature.parameters.items():
        match = re.findall(r"alias=(\w+)", str(signature.annotation))
        for alias in match:
            if alias in params_and_aliases:
                raise SyntaxError(
                    f"Multiple aliases or definitions of '{alias}' in '{func.ref}.{func.__name__}'"
                )
            else:
                if param not in func.param_aliases:
                    func.param_aliases[param] = {param}
                func.param_aliases[param].add(alias)
                params_and_aliases.add(alias)


class LoadedMod(types.ModuleType):
    """
    The LoadedMod class allows for the module loaded onto the sub to return
    custom sequencing, for instance it can be iterated over to return all
    functions
    """

    def __init__(self, name: str):
        super().__init__(name)
        self._vars = pop.data.NamespaceDict()
        self._funcs = pop.data.NamespaceDict()
        self._classes = pop.data.NamespaceDict()
        self._attrs = pop.data.MultidictCache([self._vars, self._funcs, self._classes])

    def __getattr__(self, item: str):
        try:
            return self._attrs[item]
        except KeyError as e:
            raise AttributeError(e)

    def __getitem__(self, item: str):
        return self._attrs[item]



import pop.data


async def load(
    hub: "pop.hub.Hub",
    cli: str = None,
    cli_config: dict[str, object] = None,
    config: dict[str, object] = None,
    subcommands: dict[str, object] = None,
    global_clis: list[str] = None,
    parser_args: tuple = None,
):
    """
    Use the pop-config system to load up a fresh configuration for this project
    from the included conf.py file.
    """
    if not cli_config:
        cli_config = hub._dynamic.config.get("cli_config") or {}

    # These CLI namespaces will be added on top of any cli
    if not global_clis:
        global_clis = ["log", "pop"]

    # Initialize the active cli, this is what will go into argparse
    active_cli = {}

    # Logging options and config file are part of the global namespace
    for gn in global_clis:
        active_cli.update(cli_config.get(gn, {}))

    if not subcommands:
        active_subcommands = subcommands = (
            hub._dynamic.config.get("subcommands") or {}
        ).get(cli, {})
    else:
        active_subcommands = subcommands

    # Grab the named cli last so that it can override globals
    active_cli.update(cli_config.get(cli, {}))

    # Add all the cli options to argparse and call hte parser
    cli_opts = await hub.pop.config.parse_cli(
        cli=cli,
        active_cli=active_cli,
        subcommands=active_subcommands,
        parser_args=parser_args,
    )

    # If a config file was passed in then read it
    if cli_opts and cli_opts.config:
        with open(cli_opts.config) as fh:
            config_data = hub.lib.yaml.safe_load(fh)
    else:
        config_data = {}

    # Get the plain config data that will tell us about OS vars and defaults
    if not config:
        config = hub._dynamic.config.get("config") or {}

    OPT = await hub.pop.config.prioritize(
        cli=cli, cli_opts=cli_opts, config=config, config_file_data=config_data, global_clis=global_clis
    )

    return pop.data.freeze(OPT)


async def parse_cli(
    hub,
    cli: str,
    active_cli: dict[str, object],
    subcommands: dict[str, object],
    parser_args: tuple = None,
) -> dict[str, object]:
    """
    Create a parser and parse all the CLI options.

    Args:
        hub: The POP hub instance.
        cli (str): The name of the CLI being parsed.
        active_cli (dict): The active CLI configuration.
        subcommands (dict): The subcommands configuration.

    Returns:
        argparse.Namespace: The parsed CLI options.
    """
    if not cli:
        return {}
    parser = hub.lib.argparse.ArgumentParser(description=f"{cli} CLI parser")
    subparsers = {}

    # Add subcommands to the parser
    for subcommand, opts in subcommands.items():
        subparsers[subcommand] = parser.add_subparsers(dest=subcommand, **opts)

    # Add CLI options to the parser
    groups = {}
    for name, opts in active_cli.items():
        opts = opts.copy()
        positional = opts.pop("positional", False)
        cli_name = name if positional else f"--{name.lower().replace('_', '-')}"

        options = opts.pop("options", [])
        arg_subparsers = opts.pop("subparsers", [])

        # If choices is a string then assume it is a ref on the hub and get a list of loaded mods
        choices = opts.get("choices")
        if isinstance(choices, str):
            choices = (c for c in hub[choices]._loaded.keys() if c != "init")
            opts["choices"] = choices

        # Handle argument groups
        group_name = opts.pop("group", None)
        target_parser = (
            groups.setdefault(group_name, parser.add_argument_group(group_name))
            if group_name
            else parser
        )

        if not arg_subparsers or "__global__" in arg_subparsers:
            target_parser.add_argument(cli_name, *options, **opts)
        else:
            for p in arg_subparsers:
                subparsers[p].add_argument(cli_name, *options, **opts)

    return pop.data.NamespaceDict(parser.parse_args(args=parser_args).__dict__)


PLACEHOLDER = object()


async def prioritize(
    hub,
    cli: str,
    cli_opts: dict[str, any],
    config: dict[str, any],
    config_file_data: dict[str, any],
    global_clis: list[str],
) -> pop.data.ImmutableNamespaceDict:
    """
    Prioritize configuration data from various sources.

    The order of priority is:
    1. CLI options (highest priority)
    2. Configuration file data
    3. OS environment variables
    4. Default values (lowest priority)

    Args:
        hub: The POP hub instance.
        cli (str): The name of the CLI being prioritized.
        cli_opts (dict): The parsed CLI options.
        config (dict): The configuration dictionary.
        config_file_data (dict): The data from the configuration file.

    Returns:
        pop.data.ImmutableNamespaceDict: The prioritized configuration options.
    """
    OPT = hub.lib.collections.defaultdict(dict)
    for namespace, args in config.items():
        for arg, data in args.items():
            # Initialize value to None
            value = None

            # 1. Check CLI options first
            if (namespace == cli  or namespace in global_clis) and arg in cli_opts:
                value = cli_opts.get(arg, PLACEHOLDER)

            # 2. Check config file data if CLI option is not set
            if value is None:
                value = config_file_data.get(namespace, {}).get(arg, PLACEHOLDER)

            # Skip malformed config
            if not isinstance(data, dict):
                raise TypeError("Invalid data from config.yaml: {data}")

            # 3. Check OS environment variables if config file data is not set
            if value is PLACEHOLDER and "os" in data:
                value = hub.lib.os.environ.get(arg, PLACEHOLDER)

            # 4. Use default value if none of the above are set
            if value is PLACEHOLDER:
                value = data.get("default")

            # Set the value in the OPT dictionary
            OPT[namespace][arg] = value

    return pop.data.freeze(OPT)


import argparse

argparse.ArgumentParser().parse_args



# cython: language_level=3

from collections.abc import Mapping
from cpython cimport dict
from cython cimport pint
from cpython.ref cimport PyObject


cdef frozenset DICT_ATTRS = frozenset(dir(dict))

cdef class NamespaceDict(dict[str, object]):
    def __getattr__(self, key: str):
        if key.startswith("__") and key.endswith("__"):
            self.__getattribute__(key)
        try:
            return self[key]
        except KeyError as e:
            raise AttributeError(e)

    def __setattr__(self, key: str, value: object):
        if isinstance(value, dict):
            value = NamespaceDict(value)
        self[key] = value


cdef class MultidictCache:
    cdef dict[str, object] _cache
    cdef list[dict[str, object]] _base_dicts
    cdef dict[str, list[str]] _split_cache

    def __init__(self, base_dicts=None):
        if base_dicts is None:
            base_dicts = []
        self._base_dicts = base_dicts
        self._cache = dict[str, object]()
        self._split_cache = dict[str, list[str]]()

    cpdef _clear(self):
        self._cache.clear()
        self._split_cache.clear()

    def __getitem__(self, key:str):
        if key in self._cache:
            return self._cache[key]

        if key not in self._split_cache:
            self._split_cache[key] = key.split('.')

        parts = self._split_cache[key]
        for base_dict in self._base_dicts:
            finder = base_dict
            try:
                for part in parts:
                    finder = finder[part]
                self._cache[key] = finder
                return finder
            except KeyError:
                continue

        raise KeyError(f"Item '{key}' not found in any base dictionary")

    def __iter__(self):
        return iter({k for base_dict in self._base_dicts for k in base_dict})

    def __contains__(self, item):
        return any(item in base_dict for base_dict in self._base_dicts)

cdef class ImmutableNamespaceDict:
    cdef readonly dict __data

    def __init__(self, data=None):
        if isinstance(data, ImmutableNamespaceDict):
            self.__data = data.__data
        elif isinstance(data, dict):
            self.__data = {key: freeze(value) for key, value in data.items()}
        else:
            self.__data = {}

    def __getitem__(self, key):
        try:
            return self.__data[key]
        except KeyError:
            return ImmutableNamespaceDict()

    def __getattr__(self, key):
        if key in self.__data:
            return self.__data[key]
        raise AttributeError(f"'{type(self).__name__}' object has no attribute '{key}'")

    def __setattr__(self, key, value):
        if key == '_ImmutableNamespaceDict__data':
            object.__setattr__(self, key, value)
        else:
            raise TypeError("ImmutableNamespaceDict does not support attribute assignment")

    def __setitem__(self, key, value):
        raise TypeError("ImmutableNamespaceDict does not support item assignment")

    def __call__(self):
        return unfreeze(self.__data)

    def __len__(self):
        return len(self.__data)

    def __iter__(self):
        return iter(self.__data)

    def __bool__(self):
        return bool(self.__data)

    def __contains__(self, key):
        return key in self.__data

    def __repr__(self):
        return f"{self.__data}"

    def __eq__(self, other):
        if isinstance(other, ImmutableNamespaceDict):
            return self.__data == other.__data
        elif isinstance(other, dict):
            return self.__data == other
        return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def keys(self):
        return self.__data.keys()

    def values(self):
        return self.__data.values()

    def items(self):
        return self.__data.items()

    def get(self, key, default=None):
        return self.__data.get(key, default)

    def copy(self):
        return unfreeze(self.__data)


def freeze(data):
    if isinstance(data, ImmutableNamespaceDict):
        return data
    elif isinstance(data, dict):
        return ImmutableNamespaceDict({key: freeze(value) for key, value in data.items()})
    elif isinstance(data, list):
        return tuple(freeze(value) for value in data)
    elif isinstance(data, set):
        return frozenset(freeze(value) for value in data)
    else:
        return data

def unfreeze(data):
    if isinstance(data, ImmutableNamespaceDict):
        return {key: unfreeze(value) for key, value in data.items()}
    elif isinstance(data, frozenset):
        return {unfreeze(value) for value in data}
    elif isinstance(data, tuple):
        return [unfreeze(value) for value in data]
    else:
        return data


cpdef update(dest, upd, bint recursive_update=True, bint merge_lists=False):
    """
    Recursive version of the default dict.update

    Merges upd recursively into dest

    If recursive_update=False, will use the classic dict.update, or fall back
    on a manual merge (helpful for non-dict types like FunctionWrapper)

    If merge_lists=True, will aggregate list object types instead of replace.
    The list in ``upd`` is added to the list in ``dest``, so the resulting list
    is ``dest[key] + upd[key]``. This behavior is only activated when
    recursive_update=True. By default merge_lists=False.
    """
    cdef:
        list keys = list(upd.keys())
        object key, val, dest_subkey

    if recursive_update:
        for key in keys:
            val = upd[key]
            dest_subkey = dest.get(key, None)
            if isinstance(dest_subkey, Mapping) and isinstance(val, Mapping):
                ret = update(dest_subkey, val, merge_lists=merge_lists)
                dest[key] = ret
            elif isinstance(dest_subkey, list) and isinstance(val, list) and merge_lists:
                merged = dest_subkey[:]
                merged.extend([x for x in val if x not in merged])
                dest[key] = merged
            else:
                dest[key] = upd[key]
    else:
        for key in keys:
            dest[key] = upd[key]
    return dest



"""
Find directories
"""
import importlib.resources
import os
import pathlib
import sys
from collections import defaultdict
from collections.abc import Iterable
from typing import Any

import yaml

import pop.data


def dir_list(pypath: list[str] = None, static: list[str] = None) -> list[pathlib.Path]:
    """
    Return the directories to look for modules in, pypath specifies files
    relative to an installed python package, static is for static dirs
    :param pypath: One or many python paths which will be imported
    :param static: Directories that can be explicitly passed
    """
    ret = set()
    if pypath:
        for path in pypath:
            mod = importlib.import_module(path)
            for m_path in mod.__path__:
                # If we are inside of an executable the path will be different
                ret.add(pathlib.Path(m_path))
    if static:
        ret.update(pathlib.Path(dir_) for dir_ in static)
    return sorted(ret)


def inline_dirs(dirs: Iterable[str], subdir: str) -> list[pathlib.Path]:
    """
    Look for the named subdir in the list of dirs
    :param dirs: The names of configured dynamic dirs
    :param subdir: The name of the subdir to check for in the list of dynamic dirs
    :return An extended list of dirs that includes the found subdirs
    """
    ret = set()
    for dir_ in dirs:
        check = pathlib.Path(dir_) / subdir
        if check.is_dir():
            ret.add(check)
    return sorted(ret)


def dynamic_dirs() -> dict[str, Any]:
    """
    Iterate over the available python package imports and look for configured
    dynamic dirs in pyproject.toml
    """

    len(sys.path)

    dirs = set()
    for dir_ in sys.path:
        if not dir_:
            continue
        path = pathlib.Path(dir_)
        if not path.is_dir():
            continue
        for sub in os.listdir(dir_):
            full = path / sub
            if sub.endswith(".egg-link"):
                with open(full) as rfh:
                    dirs.add(pathlib.Path(rfh.read().strip()))
            elif full.is_dir():
                dirs.add(full)

    # Set up the _dynamic return
    ret = pop.data.NamespaceDict(
        dyne=pop.data.NamespaceDict(),
        config=pop.data.NamespaceDict(),
        imports=pop.data.NamespaceDict(),
    )

    # Iterate over namespaces in sys.path
    for dir_ in dirs:
        config_file = dir_ / "config.yaml"
        try:
            if not config_file.is_file():
                continue
        except:
            continue
        dynes, configs, imports = parse_config(config_file)
        if dynes:
            pop.data.update(ret.dyne, dynes, merge_lists=True)
        if configs:
            pop.data.update(ret.config, configs, merge_lists=True)
        if imports:
            pop.data.update(ret.imports, imports)

    return ret


def parse_config(
    config_file: str,
) -> tuple[dict[str, object], dict[str, object], set[str]]:
    dyne = defaultdict(lambda: pop.data.NamespaceDict(paths=set()))
    config = pop.data.NamespaceDict(
        config=pop.data.NamespaceDict(),
        cli_config=pop.data.NamespaceDict(),
        subcommands=pop.data.NamespaceDict(),
    )
    imports = pop.data.NamespaceDict()

    config_file = pathlib.Path(config_file)
    if not config_file.is_file():
        return dyne, config, imports
    with open(config_file) as f:
        pop_config = yaml.safe_load(f) or {}

    # Gather dynamic namespace paths for this import
    for name, paths in pop_config.get("dyne", {}).items():
        for path in paths:
            ref = config_file.parent / path.replace(".", os.sep)
            dyne[name]["paths"].add(ref)

    # Get config sections
    for section in ["config", "cli_config", "subcommands"]:
        section_data = pop_config.get(section)
        if not isinstance(section_data, dict):
            continue
        for namespace, data in section_data.items():
            if data is None:
                continue
            config[section].setdefault(namespace, pop.data.NamespaceDict()).update(data)

    # Handle python imports
    for imp in pop_config.get("import", []):
        base = imp.split(".", 1)[0]
        if base not in imports:
            imports[base] = importlib.import_module(base)
        if "." in imp:
            importlib.import_module(imp)

    for name in dyne:
        dyne[name]["paths"] = sorted(dyne[name]["paths"])

    return dyne, config, imports



import inspect
from collections.abc import Callable
from collections.abc import Iterable
from typing import Any

import pop.exc
import pop.hub
import pop.loader

NO_TYPE_ANNOTATION = object()


def contract(
    raws: Iterable["pop.loader.LoadedMod"],
    mod: "pop.loader.LoadedMod",
):
    """
    Verify module level contract - functions only
    :param hub: The redistributed pop central hub
    :param raws: A list of loaded modules with contracts
    :param mod: A loader module
    """
    sig_errs = []
    sig_miss = []
    for raw in raws:
        if isinstance(raw, pop.loader.LoadError):
            sig_errs.append(str(raw))
            continue
        else:
            for fun in raw._funcs:
                if fun.startswith("sig_"):
                    tfun = fun[4:]
                    if tfun not in mod._funcs:
                        sig_miss.append(
                            f"Function '{tfun}' is missing from {mod.__file__}"
                        )
                        continue
                    sig_errs.extend(sig(mod._funcs[tfun].func, raw._funcs[fun].func))
                    sig_errs.extend(sync(mod._funcs[tfun].func, raw._funcs[fun].func))
    if sig_errs or sig_miss:
        msg = ""
        if sig_errs:
            msg += f"Signature Errors in {mod.__file__}:\n"
            for err in sig_errs:
                msg += f"{err}\n"
        if sig_miss:
            msg += f"Signature Functions Missing in {mod.__file__}:\n"
            for err in sig_miss:
                msg += f"{err}\n"
        msg = msg.strip()
        raise pop.exc.ContractSigException(msg)


def sig_map(ver: Callable) -> dict[str, Any]:
    """
    Generates the map dict for the signature verification
    """
    vsig = inspect.signature(ver)
    vparams = list(vsig.parameters.values())
    vdat = {
        "args": [],
        "v_pos": -1,
        "kw": [],
        "kwargs": False,
        "ann": {},
        "defaults": {},
    }
    for ind in range(len(vparams)):
        param = vparams[ind]
        val = param.kind.value
        name = param.name
        if val == inspect._POSITIONAL_ONLY or val == inspect._POSITIONAL_OR_KEYWORD:
            vdat["args"].append(name)
            if param.default != inspect._empty:  # Is a KW, can be inside of **kwargs
                vdat["kw"].append(name)
        elif val == inspect._VAR_POSITIONAL:
            vdat["v_pos"] = ind
        elif val == inspect._KEYWORD_ONLY:
            vdat["kw"].append(name)
        elif val == inspect._VAR_KEYWORD:
            vdat["kwargs"] = ind
        if param.annotation != inspect._empty:
            vdat["ann"][name] = param.annotation
        if param.default != inspect._empty:
            vdat["defaults"][name] = param.default
    return vdat


def sig(func: Callable, ver: Callable) -> list[str]:
    """
    Takes 2 functions, the first function is verified to have a parameter signature
    compatible with the second function
    """
    errors = []
    fsig = inspect.signature(func)
    fparams = list(fsig.parameters.values())
    vdat = sig_map(ver)
    arg_len = len(vdat["args"])
    v_pos = False
    f_name = getattr(func, "__name__", func.__class__.__name__)
    for ind in range(len(fparams)):
        param = fparams[ind]
        val = param.kind.value
        name = param.name
        has_default = param.default != inspect._empty
        ann = param.annotation

        vann = vdat["ann"].get(name, NO_TYPE_ANNOTATION)
        # Only enforce type if one is given in the signature, it won't be for *args and **kwargs
        if vann is not NO_TYPE_ANNOTATION and vann != ann:
            errors.append(f'{f_name}: Parameter "{name}" is type "{ann}" not "{vann}"')

        # if the signature argument doesn't have a default, but the function does:
        if (
            name in vdat["args"]
            and name not in vdat["defaults"]
            and param.default != inspect._empty
        ):
            errors.append(f'{f_name}: Parameter "{name}" cannot have a default value')
        # if the signature argument *does* have a default, but the function doesn't:
        elif name in vdat["defaults"] and param.default == inspect._empty:
            errors.append(f'{f_name}: Parameter "{name}" must have a default value')

        if val == inspect._POSITIONAL_ONLY or val == inspect._POSITIONAL_OR_KEYWORD:
            if ind >= arg_len:  # Past available positional args
                if not vdat["v_pos"] == -1:  # Has a *args
                    if ind >= vdat["v_pos"] and v_pos:
                        # Invalid unless it is a kw
                        if name not in vdat["kw"]:
                            # Is a kw
                            errors.append(f'Parameter "{name}" is invalid')
                        if vdat["kwargs"] is False:
                            errors.append(
                                f'{f_name}: Parameter "{name}" not defined as kw only'
                            )
                        continue
                elif vdat["kwargs"] is not False and not has_default:
                    errors.append(
                        f'{f_name}: Parameter "{name}" is past available positional params'
                    )
                elif vdat["kwargs"] is False:
                    errors.append(
                        f'{f_name}: Parameter "{name}" is past available positional params'
                    )
            else:
                v_param = vdat["args"][ind]
                if v_param != name:
                    errors.append(
                        f'{f_name}: Parameter "{name}" does not have the correct name: {v_param}'
                    )
        elif val == inspect._VAR_POSITIONAL:
            v_pos = True
            if vdat["v_pos"] == -1:
                errors.append(f"{f_name}: *args are not permitted as a parameter")
            elif ind < vdat["v_pos"]:
                errors.append(
                    f'{f_name}: Parameter "{name}" is not in the correct position for *args'
                )
        elif val == inspect._KEYWORD_ONLY:
            if name not in vdat["kw"] and not vdat["kwargs"]:
                errors.append(
                    f'{f_name}: Parameter "{name}" is not available as a kwarg'
                )
        elif val == inspect._VAR_KEYWORD:
            if vdat["kwargs"] is False:
                errors.append(f"{f_name}: **kwargs are not permitted as a parameter")
    if errors:
        if hasattr(ver.__code__, "co_filename"):
            errors.append(
                f"Enforcing signature: {ver.__code__.co_filename}::{ver.__name__}"
            )
        else:
            errors.append(f"Enforcing signature: {ver.__module__}.{ver.__name__}")
    return errors


def sync(func: Callable, ver: Callable) -> list[str]:
    """
    Compare two functions and make sure that they match in being synchronous or asynchronous
    """
    errors = []
    f_name = getattr(func, "__name__", func.__class__.__name__)

    f_type = func_type(func)
    v_type = func_type(ver)

    if "async" in v_type and "async" not in f_type:
        errors.append(f"{f_name}: Function must be asynchronous")
    if "gen" in v_type and "gen" not in f_type:
        errors.append(
            f"{f_name}: Function does not yield, generator required by signature"
        )

    return errors


def func_type(func: Callable) -> str:
    if inspect.iscoroutinefunction(func):
        return "asynchronous"
    elif inspect.isasyncgenfunction(func):
        return "asynchronous generator"
    elif inspect.isgenerator(func):
        return "generator"
    else:
        return "synchronous"



async def __init__(hub):
    hub.log.LOGGER = {}
    hub.log.FILE_HANDLER = None
    hub.log.STREAM_HANDLER = None
    hub.log.LEVEL = {
        "notset": hub.lib.logging.NOTSET,
        "trace": 5,
        "debug": hub.lib.logging.DEBUG,
        "info": hub.lib.logging.INFO,
        "warn": hub.lib.logging.WARN,
        "warning": hub.lib.logging.WARNING,
        "error": hub.lib.logging.ERROR,
        "fatal": hub.lib.logging.FATAL,
        "critical": hub.lib.logging.CRITICAL,
    }



"""
Contracts to enforce loader objects
"""
import functools
import inspect
from collections.abc import Iterable
from types import ModuleType
from typing import get_args
from typing import get_origin
from typing import get_type_hints

import pop.exc
import pop.hub
import pop.verify

try:
    # Only do strict type checking in a dev environment
    import pytest  # noqa

    STRICT_TYPE_CHECKING: bool = True
except ImportError:
    STRICT_TYPE_CHECKING: bool = False


class ContractedContext:
    """
    Contracted function calling context
    """

    def __init__(
        self,
        func: functools.partial,
        args: Iterable,
        kwargs: dict,
        signature,
        ref,
        name,
        ret=None,
        cache=None,
    ):  # pylint: disable=too-many-arguments
        if cache is None:
            cache = {}

        self.func = func
        self.args = list(args)
        self.kwargs = kwargs
        self.signature = signature
        self.ref = ref
        self.__name__ = name
        self.ret = ret
        self.cache = cache

    def get_arguments(self):
        """
        Return a dictionary of all arguments that will be passed to the function and their
        values, including default arguments.
        """
        if "__bound_signature__" not in self.cache:
            self.cache["__bound_signature__"] = self.signature.bind(
                *self.args, **self.kwargs
            )
            # Apply any default values from the signature
            self.cache["__bound_signature__"].apply_defaults()
        return self.cache["__bound_signature__"].arguments


def load_contract(
    contracts: "pop.hub.Sub",
    default_contracts: list[str],
    mod: ModuleType,
    name: str,
) -> list["pop.loader.LoadedMod"]:
    """
    return a Contract object loaded up
    Dynamically create the correct Contracted type
    :param contracts: Contracts functions to add to the sub
    :param default_contracts: The contracts that have been marked as defaults
    :param mod: A loader module
    :param name: The name of the module to get from the loader
    """
    raws = []
    loaded_contracts = []
    if hasattr(contracts, name):
        loaded_contracts.append(name)
        raws.append(getattr(contracts, name))
    if hasattr(contracts, "init"):
        if "init" not in loaded_contracts:
            loaded_contracts.append("init")
            raws.append(getattr(contracts, "init"))
    if hasattr(mod, "__contracts__"):
        cnames = getattr(mod, "__contracts__")
        for cname in cnames:
            if cname in contracts:
                loaded_contracts.append(cname)
                raws.append(getattr(contracts, cname))
    return raws


class Contracted:
    """
    This class wraps functions that have a contract associated with them
    and executes the contract routines
    """

    def __init__(
        self,
        hub: "pop.hub.Hub",
        contracts: list["pop.hub.Sub"],
        func: functools.partial,
        ref: str,
        name: str,
        implicit_hub: bool = True,
        param_aliases: dict = None,
    ):
        """
        :param hub: The redistributed pop central hub
        :param contracts: Contracts functions to add to the sub
        :param func: The contracted function to call
        :param ref: The reference to the function on the hub
        :param name: An alias for the function
        :param implicit_hub: True if a hub should be implicitly injected into the "call" method
        :param param_aliases: A dictionary that maps parameters to a set of its aliases
        """
        self.__dict__.update(
            getattr(func, "__dict__", {})
        )  # do this first so we later overwrite any conflicts
        self.func = func
        self.ref = ref
        self.__name__ = name
        self.signature = inspect.signature(self.func)
        self._sig_errors = []
        self.__wrapped__ = func
        self.hub = hub
        self.contracts = contracts or []
        self.implicit_hub = implicit_hub
        self._load_contracts()
        self.param_aliases = param_aliases or {}

    def _get_contracts_by_type(self, contract_type: str = "pre") -> list["Contracted"]:
        """
        :param contract_type: One of "call", "pre", "post", or "sig"
        """
        matches = []
        fn_contract_name = f"{contract_type}_{self.__name__}"
        for contract in self.contracts:
            if hasattr(contract, fn_contract_name):
                matches.append(getattr(contract, fn_contract_name))
            if hasattr(contract, contract_type):
                matches.append(getattr(contract, contract_type))

        if contract_type == "post":
            matches.reverse()

        return matches

    def _load_contracts(self):
        # if Contracted - only allow regular pre/post
        # if ContractedAsync - allow coroutines and functions

        self.contract_functions = {
            "pre": self._get_contracts_by_type("pre"),
            "call": self._get_contracts_by_type("call")[:1],
            "post": self._get_contracts_by_type("post"),
        }
        self._has_contracts = sum(len(l) for l in self.contract_functions.values()) > 0

    async def __call__(self, *args, **kwargs):
        if self.implicit_hub:
            args = (self.hub,) + args

        if STRICT_TYPE_CHECKING:
            assert strict_type_checking(self, *args, **kwargs)

        if not self._has_contracts:
            ret = await self.func(*args, **kwargs)
            if STRICT_TYPE_CHECKING:
                assert strict_return_checking(self, ret)
            return ret
        contract_context = ContractedContext(
            self.func, args, kwargs, self.signature, self.ref, self.__name__
        )

        # Process pre contracts
        for fn in self.contract_functions["pre"]:
            pre_ret = await fn(contract_context)

            await self.hub.pop.contract.process_pre_result(pre_ret, fn, self)

        # Call the one call contract
        if self.contract_functions["call"]:
            ret = await self.contract_functions["call"][0](contract_context)
        else:
            ret = await self.func(*contract_context.args, **contract_context.kwargs)

        # Handle post contracts
        for fn in self.contract_functions["post"]:
            contract_context.ret = ret
            post_ret = await fn(contract_context)
            if post_ret is not None:
                ret = post_ret

        if STRICT_TYPE_CHECKING:
            assert strict_return_checking(self, ret)

        return ret

    def __getstate__(self):
        return dict(
            hub=self.hub,
            ref=self.ref,
            name=self.__name__,
            implicit_hub=self.implicit_hub,
            contracts=self.contracts,
            param_aliases=self.param_aliases,
        )

    def __setstate__(self, state):
        hub = state["hub"]
        ref = state["ref"]
        name = state["name"]
        func = hub[ref][name].func
        self.__init__(
            hub=hub,
            func=func,
            ref=ref,
            name=name,
            implicit_hub=state["implicit_hub"],
            contracts=state["contracts"],
            param_aliases=state["param_aliases"],
        )


class ContractedAsyncGen(Contracted):
    async def __call__(self, *args, **kwargs):
        if self.implicit_hub:
            args = (self.hub,) + args

        if STRICT_TYPE_CHECKING:
            assert strict_type_checking(self, *args, **kwargs)

        if not self._has_contracts:
            async for chunk in self.func(*args, **kwargs):
                yield chunk
            return
        contract_context = ContractedContext(
            self.func, args, kwargs, self.signature, self.ref, self.__name__
        )

        for fn in self.contract_functions["pre"]:
            pre_ret = await fn(contract_context)

            await self.hub.pop.contract.process_pre_result(pre_ret, fn, self)
        chunk = None
        if self.contract_functions["call"]:
            async for chunk in self.contract_functions["call"][0](contract_context):
                yield chunk
        else:
            async for chunk in self.func(
                *contract_context.args, **contract_context.kwargs
            ):
                yield chunk
        ret = chunk
        for fn in self.contract_functions["post"]:
            contract_context.ret = ret
            post_ret = await fn(contract_context)
            if post_ret is not None:
                ret = post_ret

        if STRICT_TYPE_CHECKING:
            assert strict_return_checking(self, ret)


def create_contracted(
    hub: "pop.hub.Hub",
    contracts: list["pop.loader.LoadedMod"],
    func: functools.partial,
    ref: str,
    name: str,
    implicit_hub: bool = True,
) -> Contracted:
    """
    Dynamically create the correct Contracted type
    :param hub: The redistributed pop central hub
    :param contracts: Contracts functions to add to the sub
    :param func: The contracted function to call
    :param ref: The reference to the function on the hub
    :param name: The name of the module to get from the loader
    :param implicit_hub: True if a hub should be implicitly injected into the "call" method
    """
    if inspect.isasyncgenfunction(func):
        return ContractedAsyncGen(hub, contracts, func, ref, name, implicit_hub)
    else:
        return Contracted(hub, contracts, func, ref, name, implicit_hub)


def strict_type_checking(contract: Contracted, *args, **kwargs) -> bool:
    # Perform type checking based on the signature and type hints
    bound_args = contract.signature.bind(*args, **kwargs)
    type_hints = get_type_hints(contract.func)

    for param_name, value in bound_args.arguments.items():
        expected_type = type_hints.get(param_name)
        if expected_type:
            origin = get_origin(expected_type)
            type_args = get_args(expected_type)

            if origin and not isinstance(value, origin):
                raise TypeError(
                    f"Argument '{param_name}' must be of type {origin.__name__}, got {type(value).__name__}"
                )

            if type_args:
                # Add checks for specific container types if needed
                pass

    return True


def strict_return_checking(contract: Contracted, ret: any) -> bool:
    type_hints = get_type_hints(contract.func)
    expected_return_type = type_hints.get("return")

    if expected_return_type:
        origin = get_origin(expected_return_type)
        type_args = get_args(expected_return_type)

        if origin and not isinstance(ret, origin):
            raise TypeError(
                f"Return value must be of type {origin.__name__}, got {type(ret).__name__}"
            )

        if type_args:
            # Add checks for specific container types if needed
            pass

    return True



from typing import Any


async def get(hub: "pop.hub.Hub") -> dict[str, Any]:
    """
    Retrive the dynamic dirs data for this hub, if dynamic dirs have not been
    gathered yet then gather it.
    """
    return hub._dynamic.dyne
